#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Aug  5 09:53:01 2022

@author: scatr
"""

from sklearn.linear_model import ridge_regression
from sklearn.linear_model import LinearRegression
from itertools import combinations
from inspect import signature
import numpy as np
import pysindy as ps

def _custom_lib(X, library_functions):
    """

    Parameters
    ----------
    X : np.array
        The input trajectory data, [time, state].
    library_functions : list
        List of user-specified lambda functions from which to build the fitting library.

    Returns
    -------
    Theta : np.array
        The library of features (eg, polynomial terms) to be fitted to [time, n_features].

    """
    
    n_states = X.shape[-1] #number of equations 
    n_time = X.shape[0]
    

    total_combs = 0 
    for ii, func in enumerate(library_functions):
        #find the total combinations in the library to allocate the memory and avoid appends
        sig=signature(func)
        arg_nos = len(sig.parameters)
        
        total_combs+= len(list(combinations(range(n_states), arg_nos))) 

    Theta = np.empty((n_time, total_combs))
    
    
    theta_ind = 0     
    #build the theta lbrary
    for ii, func in enumerate(library_functions):
        
        sig=signature(func)
        arg_nos = len(sig.parameters)
        
        #find the combinations required for the custom function
        combs = list(combinations(range(n_states), arg_nos))
        
        for c in combs:
            Theta[:,theta_ind] = func(*[X[:,i] for i in c])
            theta_ind+=1
            
    return Theta


def _fix_coefficients(fixed_coefficients, xder, Theta):
    """
    This function will take coefficients that are to be fixed in the library
        \dot{f} - M = \theta X 
    Where M are the known coefficients and those terms are no longer included in the library 
    """
    
    new_xder = xder - np.dot(Theta, fixed_coefficients)
    
    return new_xder


def _initial_guess_least_squares(Theta, xder, Theta_0):
    """
    Minimizes the least squares solution with an additional term in the objective function
    related to the initial guess of the coefficients. This is the same solution as the 
    regularised least squares Tikhonov. 

    Parameters
    ----------
    Theta : np.array
        The feature library.
    xder : TYPE
        DESCRIPTION.
    Theta_0 : TYPE
        DESCRIPTION.

    Returns
    -------
    TYPE
        DESCRIPTION.

    """
    
    return (np.linalg.inv(Theta.T@Theta+np.identity(Theta.shape[0])))@(Theta.T@xder+Theta_0)
        

def get_coef(xder, Theta, library_function_names, coef=None, initial_guess=None, unbias=True, maxiter=10, thresholds = None,
            alpha = 0.01):
    # algorithm as in Brunton et al 2016
    """
     Parameters
     ----------
     xint : np.array
         Input trajectory data [time, n_states].
     time : np.array
         Time range of trajectory data.
     library_functions : list
         List of user-specified lambda functions from .
     alpha : float, optional
         Regularisation parameter used in the ridge regression. 
         The default is 0.01.
     constrain_library : np.array, optional
         Array of boolean specifying which terms to regress over [n_states, n_features]. 
         The default is None.
     initial_guess: np.array, optional
         An estimate of the coefficients to be used in the optimisation proceedure
     threshold : float, optional
         Hard threshold to cut off coefficients. The default is 0.1.
     maxiter : int, optional
         Maximum number of iterations of the STLSQ algorithm. The default is 20.
     unbias : bool, optional
         Perform an unregularised regression in the last step. The default is True.
     thresholds : list, optional
         Separate thresholds for each state. The default is None.
    
     Returns
     -------
     Xi : np.array
         Model coefficients.
     coefhist : list
         List of coefficients history from the iterative proceedure.
    """        
    if thresholds is None:
        raise ValueError("Thresholds must be assigned as a array equal to the number of targets")

    #code which removes terms in already constrained terms from the library.
    
    Xi = np.zeros((Theta.shape[-1], xder.shape[-1]))
    initial_inds = np.ones(Xi.shape).astype(int)   


    #initial regularised least squares guess
    if initial_guess is None:
        for n in range(xder.shape[-1]):
            #Xi[initial_inds[n],n] = LinearRegression().fit(Theta[..., initial_inds[n]], xder[:,n]).coef_
            Xi[initial_inds[n], n] = ridge_regression(Theta[..., initial_inds[n,:]], xder[:, n], alpha)
    else:
        print("Starting from initial guess")
        assert Xi.shape == initial_guess.shape, "The passed coefficient shapes do not match"
        
        Xi = _initial_guess_least_squares(Theta, xder, initial_guess)
        
                
    dim = xder.shape[-1]
    
    
    #create the coefficient store
    if unbias==True:
        coefhist = np.zeros((maxiter+2, Xi.shape[1], Xi.shape[0]))
    else:
        coefhist = np.zeros((maxiter+1, Xi.shape[1], Xi.shape[0]))
    coefhist[0,:,:] = Xi.T
        
    #################################### THRESHOLDING ALGORITHM ######################################
    
    for k in range(maxiter):
        
        # do separate fitting for each variable - I guess otherwise the matrices Theta and xder can't be put together
        for ind in range(xder.shape[-1]):
            smallinds = (np.abs(Xi[:,ind]) <= thresholds[ind]) # find where in Xi coeffs are smaller than threshold and zero them
            
            Xi[smallinds, ind]=0.0
            
            if np.count_nonzero(Xi[:,ind]==0)==Theta.shape[-1]:
                print("All coefficients eliminated")
                break
                
            biginds = ~smallinds  # indices of coeffs that are not small by our definition
            if initial_guess is None:
                Xi[biginds, ind] = ridge_regression(Theta[..., biginds], xder[:, ind], alpha) #LinearRegression().fit(Theta[..., biginds], xder[:, ind]).coef_.T # new regression
            else:
                Xi[biginds, ind] =  _initial_guess_least_squares(Theta[..., biginds], xder[:, ind], initial_guess) #not sure this makes sense
        coefhist[k+1,:,:] = Xi.T
        
        if np.array_equal(np.where(coefhist[k,:,:]), np.where(coefhist[k+1,:,:])):
            print("No change in regression coefficients")
            coefhist = coefhist[:k+3,:,:] #remove the bits that we no longer 
            break
        
        
    if unbias == True:
        #perform a final regression
        for n in range(xder.shape[-1]):
            biginds = np.abs(Xi[:,n]) > 0
            if np.any(biginds):
                Xi[biginds, n] = LinearRegression().fit(Theta[..., biginds], xder[:,n]).coef_
            
                
        coefhist[-1,:,:] = Xi.T


    ## hijack the pySINDy class for printing purposes
    #feature_library = ps.CustomLibrary(library_functions=library_functions, function_names=library_function_names)
    #optimizer =  ps.STLSQ(threshold=threshold, alpha=alpha,
    #                         thresholds=thresholds)
    
    #odel = ps.SINDy(feature_library=feature_library, optimizer=optimizer, feature_names=feature_names)
    #model.fit(xint, time, multiple_trajectories = True) # I have to fit the model to instantiate the class

    model_coefs = coefhist[-1,:,:]
    
    return Xi, coefhist, model_coefs
