"""
Build python script to reproduce Garaud et al. 2010

Usage:
      python garaud_sindy_1.py

Options:

"""
import h5py
import numpy as np
import matplotlib
#matplotlib.use('Agg')
import matplotlib.pyplot as plt
#from dedalus.extras import plot_tools
from mpl_toolkits.mplot3d import Axes3D
from sklearn.linear_model import Lasso
from scipy.io import loadmat
from sklearn.metrics import mean_squared_error
from scipy.integrate import solve_ivp
from array import array
from STLSQ_functions import *
from scipy.interpolate import interp1d 
from scipy.interpolate import interp2d 
from scipy.interpolate import interpn 

import pysindy as ps

# Ignore matplotlib deprecation warnings
import warnings
warnings.filterwarnings("ignore", category=UserWarning)

"""
Points with the code:
    1. We should try to understand how the size of the integration domain impacts the results. 
What is important in smaller domains vs bigger domains? Is a larger domain integral always better? Is it only better for large scale terms?
    2. How do we choose the time averaging, what if we do almost no time averaging? 
    3. Why is the equation in W not working as well? 
    4. How will we identify the viscous terms? I think either normalisations or a different optimiser
    5. If we get this to work, can we find the mean terms and remove them in the identification process? is this possible?
    6. Why are we not seeing viscous dissipation terms in the temperature equation even though they should be order 1?
    7. This problem seems to perform poorly with high L2 regularisation for some reason 

    8. It would be great if we could justify the flow being incompressible and thus not needing a pressure evolution equation. 
"""

# Seed the random number generators for reproducibility
np.random.seed(100)

integrator_keywords = {}
integrator_keywords['rtol'] = 1e-12
integrator_keywords['method'] = 'LSODA'
integrator_keywords['atol'] = 1e-12

####
file_range = [1] # seems like 128x128x192 can only be processed as a single file
#file_range = [1,2]
#file_range = [1,2,3,4,5]
#file_range = [1,2,3,4,5,6,7,8]
####
time = np.array([]) # read in the simulation time
#nx, ny, nz = 64, 64, 96
nx, ny, nz = 128, 128, 192
nt = int(50*file_range[-1])

Pres_store = np.zeros((nt, nx, ny, nz), dtype=np.float64)
Temp_store = np.zeros((nt, nx, ny, nz), dtype=np.float64)
U_store = np.zeros((nt, nx, ny, nz), dtype=np.float64)
V_store = np.zeros((nt, nx, ny, nz), dtype=np.float64)
W_store = np.zeros((nt, nx, ny, nz), dtype=np.float64)

for qq, ii in enumerate(file_range):
### Ra 1e4 ###
#    with h5py.File('/nobackup/matcjw/dedalus/3dconvection/F-C-Ra-1e4/restart/snapshots/snapshots_s{}.h5'.format(ii)) as file:
#    with h5py.File('/nobackup/matcjw/dedalus/3dconvection/F-C-Ra-1e4/lowres/restart/snapshots/snapshots_s{}.h5'.format(ii)) as file:
### Ra 1e5 ###
#    with h5py.File('/nobackup/matcjw/dedalus/3dconvection/F-C-Ra-1e5/restart/snapshots/snapshots_s{}.h5'.format(ii)) as file:
#    with h5py.File('/nobackup/matcjw/dedalus/3dconvection/F-C-Ra-1e5/lowres/restart/snapshots/snapshots_s{}.h5'.format(ii)) as file:
### Ra 1e6 ###
    with h5py.File('/nobackup/matcjw/dedalus/3dconvection/F-C-Ra-1e6/restart/snapshots/snapshots_s{}.h5'.format(ii)) as file:
#    with h5py.File('/nobackup/matcjw/dedalus/3dconvection/F-C-Ra-1e6/lowres/restart/snapshots/snapshots_s{}.h5'.format(ii)) as file:
        print("Doing file", ii)
        if ii==1:
            print(file["scales"].keys())
            xdomain = np.array(file["tasks/pressure"].dims[1][0])
            ydomain = np.array(file["tasks/pressure"].dims[2][0])
            zdomain = np.array(file["tasks/pressure"].dims[3][0])
        # Time stamps
        time = np.append(time, file["scales/sim_time"][:])
        
        print(len(time))
        # Full data
        if qq+50<nt:
            Pres_store[qq*50:(qq+1)*50,:,:] = file['tasks/pressure'][:,:,:]  # Pressure
            Temp_store[qq*50:(qq+1)*50,:,:] = file['tasks/buoyancy'][:,:,:]  # Temperature / buoyancy
            U_store[qq*50:(qq+1)*50,:,:] = file['tasks/u_x'][:,:,:]          # u_x; x-velocity
            V_store[qq*50:(qq+1)*50,:,:] = file['tasks/u_y'][:,:,:]          # u_y; y-velocity
            W_store[qq*50:(qq+1)*50,:,:] = file['tasks/u_z'][:,:,:]          # u_z; z-velocity
        else:
            Pres_store[qq*50:,:,:] = file['tasks/pressure'][:,:,:]  # Pressure
            Temp_store[qq*50:,:,:] = file['tasks/buoyancy'][:,:,:]  # Temperature / buoyancy
            U_store[qq*50:,:,:] = file['tasks/u_x'][:,:,:]          # u_x; x-velocity
            V_store[qq*50:,:,:] = file['tasks/u_y'][:,:,:]          # u_y; y-velocity
            W_store[qq*50:,:,:] = file['tasks/u_z'][:,:,:]          # u_z; z-velocity

# Transpose axes into what SINDy expects (time,x,y,z) -> (x,y,z,time)
Pres = np.transpose(Pres_store, (1,2,3,0))
Temp = np.transpose(Temp_store, (1,2,3,0))
U = np.transpose(U_store, (1,2,3,0))
V = np.transpose(V_store, (1,2,3,0))
W = np.transpose(W_store, (1,2,3,0))

del Pres_store, Temp_store, U_store, V_store, W_store

# Note: we now have the necessary data in the right form for PDEFIND vis SINDy

# Print confirmation data
print('Temp has shape ',Temp.shape,' and size ',Temp.size)

# interpolation from Chebyshev to Fourier
print(len(time))
#new_z = np.linspace(0,1,nz+1)[:-1] # Uniform z space
new_z = np.linspace(0.5*(1/nz),1+0.5*(1/nz),nz+1)[:-1] # Uniform z space
new_Pres = np.zeros((nx, ny, nz, nt), dtype=np.float64)
new_Temp = np.zeros((nx, ny, nz, nt), dtype=np.float64)
new_U = np.zeros((nx, ny, nz, nt), dtype=np.float64)
new_V = np.zeros((nx, ny, nz, nt), dtype=np.float64)
new_W = np.zeros((nx, ny, nz, nt), dtype=np.float64)
print('New Temp has shape ',new_Temp.shape,' and size ',new_Temp.size)

for jj in range(nx):
    for kk in range(ny):
        for ii in range(len(time)):
            # interpolate on z Chebyshev to uniform spacing
            interp_p = interp1d(zdomain,Pres[jj,kk,:,ii],kind='cubic')
            new_Pres[jj,kk,:,ii] = interp_p(new_z)
            interp_t = interp1d(zdomain,Temp[jj,kk,:,ii],kind='cubic')
            new_Temp[jj,kk,:,ii] = interp_t(new_z)
            interp_U = interp1d(zdomain,U[jj,kk,:,ii],kind='cubic')
            new_U[jj,kk,:,ii] = interp_U(new_z)
            interp_V = interp1d(zdomain,V[jj,kk,:,ii],kind='cubic')
            new_V[jj,kk,:,ii] = interp_V(new_z)
            interp_W = interp1d(zdomain,W[jj,kk,:,ii],kind='cubic')
            new_W[jj,kk,:,ii] = interp_W(new_z)

#print(xdomain)
#print(ydomain)
#print(zdomain)
#print(new_z)
print('Interpolated New Temp has shape ',new_Temp.shape,' and size ',new_Temp.size)

#print(new_Pres.shape)
#print(new_Temp.shape)
#print(new_U.shape)
#print(new_V.shape)
#print(new_W.shape)

# Test - interpolation results seems OK!
#print('Old: ',xdomain[int(nx/2)],ydomain[int(ny/2)],zdomain[int((nz/2)+1)],Pres[int(nx/2),int(ny/2),int((nz/2)+1),5])
#print('Old: ',xdomain[int(nx/2)],ydomain[int(ny/2)],zdomain[int(nz/2)],Pres[int(nx/2),int(ny/2),int(nz/2),5])
#print('Old: ',xdomain[int(nx/2)],ydomain[int(ny/2)],zdomain[int((nz/2)-1)],Pres[int(nx/2),int(ny/2),int((nz/2)-1),5])
#print('Old: ',xdomain[int(nx/2)],ydomain[int(ny/2)],zdomain[int((nz/2)-2)],Pres[int(nx/2),int(ny/2),int((nz/2)-2),5])
#print('New: ',xdomain[int(nx/2)],ydomain[int(ny/2)],new_z[int(nz/2)],new_Pres[int(nx/2),int(ny/2),int(nz/2),5])

#exit()

# Use interpolated data
Pres = new_Pres
Temp = new_Temp
U = new_U
V = new_V
W = new_W

# Use new uniform z domain
zdomain = new_z

#%% I'll try to keep the convention similar to the compressible example (this isn't that much data)
#u_shaped_noiseless = np.zeros((Ubar.shape[0], Ubar.shape[1], 7)) # 1D z,t

u_shaped_noiseless = np.zeros((U.shape[0], U.shape[1], U.shape[2], U.shape[3], 5))
u_shaped_noiseless[:,:,:,:,0] = U
u_shaped_noiseless[:,:,:,:,1] = V
u_shaped_noiseless[:,:,:,:,2] = W
u_shaped_noiseless[:,:,:,:,3] = Pres
u_shaped_noiseless[:,:,:,:,4] = Temp

# the control variables are essentially terms we want included in the library but do not have independent equations

print("u data shape", u_shaped_noiseless.shape)
truncate_time = None #index to truncate at

#u_shaped_noiseless = u_shaped_noiseless[:,:truncate_time, :] #truncate the data to make it slightly easier to work with 
u_shaped_noiseless = u_shaped_noiseless[:,:,:,:truncate_time, :] #truncate the data to make it slightly easier to work with 

time = time[:truncate_time]
print(u_shaped_noiseless.shape)

######## truncate spatially ##########
x_1, x_2 = 0, None
xdomain = xdomain[x_1:x_2]
u_shaped_noiseless = u_shaped_noiseless[x_1:x_2,:,:,:,:] 

y_1, y_2 = 0, None
ydomain = ydomain[y_1:y_2]
u_shaped_noiseless = u_shaped_noiseless[:,y_1:y_2,:,:,:]

z_1, z_2 = 0, None
zdomain = zdomain[z_1:z_2]
u_shaped_noiseless = u_shaped_noiseless[:,:,z_1:z_2,:,:]

print("u shaped noiseless", u_shaped_noiseless.shape)
#####################################

nt = u_shaped_noiseless.shape[3] # (nx,ny,nz,time,variable)
t = time #change to the output of Dedalus
t -= t[0] #change to be from zero
start_time, end_time = t[0], t[-1]
print('Time:')
print(time[0])
print(time[-1])
#print(time)

# X grid. Fourier grid in realspace linearly spaced from zero.
nx = u_shaped_noiseless.shape[0] # (nx,nz,time,variable)
Lx0 = xdomain[0]
Lx = xdomain[-1] 
x = xdomain #np.linspace(0, Lx, nx, endpoint=False)
#dx = x[1] - x[0]
print('X-grid:')
print(x[0])
print(x[-1])
#print(x)
#print(x.shape)
print(x.size)

# Y grid. Fourier grid in realspace linearly spaced from zero.
ny = u_shaped_noiseless.shape[1] # (nx,ny,nz,time,variable)
Ly0 = ydomain[0]
Ly = ydomain[-1] 
y = ydomain #np.linspace(0, Ly, ny, endpoint=False)
#dy = y[1] - y[0]
print('Y-grid:')
print(y[0])
print(y[-1])
#print(y)
#print(y.shape)
print(y.size)

# Z grid. Fourier grid in realspace linearly spaced from zero.
nz = u_shaped_noiseless.shape[2] # (nx,nz,time,variable)
Lz0 = zdomain[0]
Lz = zdomain[-1] # Full domain was 2, but we removed haslf the data at read in
z = zdomain[:u_shaped_noiseless.shape[2]] #index the original z domain to the truncation
#dz = z[1] - z[0]
print('Z-grid:')
print(z[0])
print(z[-1])
#print(z)
#print(z.shape)
print(z.size)

# Question: are these grid points the left edge of cells? DO we want centre of cells (add dx/z or dz/2)?

#
spatiotemporal_grid = np.zeros((nx, ny, nz,  nt, 4))

spatiotemporal_grid[:,:,:,:,0] = x[:,np.newaxis,np.newaxis,np.newaxis] 
spatiotemporal_grid[:,:,:,:,1] = y[np.newaxis,:,np.newaxis,np.newaxis]
spatiotemporal_grid[:,:,:,:,2] = z[np.newaxis,np.newaxis,:,np.newaxis]
spatiotemporal_grid[:,:,:,:,3] = t[np.newaxis,np.newaxis,np.newaxis,:]

#X, T = np.meshgrid(z, t) # UPDATE TO 3D?
XT = spatiotemporal_grid # UPDATE TO 3D?

######################### GENERAL PDE LIBRARY TERMS ########################################
divisions_x = 48
divisions_y = 48
divisions_z = 10
divisions_t = 8
#divisions_t = 8

K = 300

weak_lin_lib = ps.WeakPDELibrary(
    library_functions=[lambda x: x],
    function_names = [lambda x: x],
    derivative_order=2, 
    spatiotemporal_grid=XT,
    include_interaction = True,
    periodic = False,   
    p=6, # Our highest order term is second derivatives so we need to make sure this is set right
    #uniform=False, there's some argument we should check is false
    K=K, #number of sub-domains to evaluate the integral on
    H_xt=[(Lx-Lx0) / divisions_x, (Ly-Ly0) / divisions_y, (Lz-Lz0) / divisions_z, (end_time-start_time) / divisions_t]
)

print("Created the weak pde library") 

#exit()

########## Quick and dirty test to obtain library structure #########
#n_coeff=28
#opt = ps.MIOSR(target_sparsity = n_coeff, alpha=0.00000000000000001) #works with small alpha
#model = ps.SINDy(feature_library=weak_lin_lib, optimizer=opt, feature_names = ["U", "V", "W", "P", "T"])
#model.fit(u_shaped_noiseless, t=t, unbias=False) #set unbias = False with MIOSR
#model.print(precision=8)
#model.get_feature_names() 
#print(model.get_feature_names())
#exit()
#
################ Constrain each equation so only the correct terms appear ############
"""
The constraints have to be of the size (n_constraints, n_features*n_targets)
this means we have to have a row for every term for every each in the system
In our case n_constraints will be the sum of the number of terms in the feature 
library - the number of terms with unknown coefficients for each equation
"""
lib_terms = ['U', 'V', 'W', 'P', 'T', 'U_3', 'V_3', 'W_3', 'P_3', 'T_3', 
             'U_33', 'V_33', 'W_33', 'P_33', 'T_33', 'U_2', 'V_2', 'W_2', 'P_2', 'T_2', 
             'U_23', 'V_23', 'W_23', 'P_23', 'T_23', 'U_22', 'V_22', 'W_22', 'P_22', 'T_22', 
             'U_1', 'V_1', 'W_1', 'P_1', 'T_1', 'U_13', 'V_13', 'W_13', 'P_13', 'T_13', 
             'U_12', 'V_12', 'W_12', 'P_12', 'T_12', 'U_11', 'V_11', 'W_11', 'P_11', 'T_11', 
             'UU_3', 'VU_3', 'WU_3', 'PU_3', 'TU_3', 'UV_3', 'VV_3', 'WV_3', 'PV_3', 'TV_3', 
             'UW_3', 'VW_3', 'WW_3', 'PW_3', 'TW_3', 'UP_3', 'VP_3', 'WP_3', 'PP_3', 'TP_3', 
             'UT_3', 'VT_3', 'WT_3', 'PT_3', 'TT_3', 
             'UU_33', 'VU_33', 'WU_33', 'PU_33', 'TU_33', 'UV_33', 'VV_33', 'WV_33', 'PV_33', 'TV_33', 
             'UW_33', 'VW_33', 'WW_33', 'PW_33', 'TW_33', 'UP_33', 'VP_33', 'WP_33', 'PP_33', 'TP_33', 
             'UT_33', 'VT_33', 'WT_33', 'PT_33', 'TT_33', 
             'UU_2', 'VU_2', 'WU_2', 'PU_2', 'TU_2', 'UV_2', 'VV_2', 'WV_2', 'PV_2', 'TV_2', 
             'UW_2', 'VW_2', 'WW_2', 'PW_2', 'TW_2', 'UP_2', 'VP_2', 'WP_2', 'PP_2', 'TP_2', 
             'UT_2', 'VT_2', 'WT_2', 'PT_2', 'TT_2', 
             'UU_23', 'VU_23', 'WU_23', 'PU_23', 'TU_23', 'UV_23', 'VV_23', 'WV_23', 'PV_23', 'TV_23', 
             'UW_23', 'VW_23', 'WW_23', 'PW_23', 'TW_23', 'UP_23', 'VP_23', 'WP_23', 'PP_23', 'TP_23', 
             'UT_23', 'VT_23', 'WT_23', 'PT_23', 'TT_23', 
             'UU_22', 'VU_22', 'WU_22', 'PU_22', 'TU_22', 'UV_22', 'VV_22', 'WV_22', 'PV_22', 'TV_22', 
             'UW_22', 'VW_22', 'WW_22', 'PW_22', 'TW_22', 'UP_22', 'VP_22', 'WP_22', 'PP_22', 'TP_22', 
             'UT_22', 'VT_22', 'WT_22', 'PT_22', 'TT_22', 
             'UU_1', 'VU_1', 'WU_1', 'PU_1', 'TU_1', 'UV_1', 'VV_1', 'WV_1', 'PV_1', 'TV_1', 
             'UW_1', 'VW_1', 'WW_1', 'PW_1', 'TW_1', 'UP_1', 'VP_1', 'WP_1', 'PP_1', 'TP_1', 
             'UT_1', 'VT_1', 'WT_1', 'PT_1', 'TT_1', 
             'UU_13', 'VU_13', 'WU_13', 'PU_13', 'TU_13', 'UV_13', 'VV_13', 'WV_13', 'PV_13', 'TV_13', 
             'UW_13', 'VW_13', 'WW_13', 'PW_13', 'TW_13', 'UP_13', 'VP_13', 'WP_13', 'PP_13', 'TP_13', 
             'UT_13', 'VT_13', 'WT_13', 'PT_13', 'TT_13', 
             'UU_12', 'VU_12', 'WU_12', 'PU_12', 'TU_12', 'UV_12', 'VV_12', 'WV_12', 'PV_12', 'TV_12', 
             'UW_12', 'VW_12', 'WW_12', 'PW_12', 'TW_12', 'UP_12', 'VP_12', 'WP_12', 'PP_12', 'TP_12', 
             'UT_12', 'VT_12', 'WT_12', 'PT_12', 'TT_12', 
             'UU_11', 'VU_11', 'WU_11', 'PU_11', 'TU_11', 'UV_11', 'VV_11', 'WV_11', 'PV_11', 'TV_11', 
             'UW_11', 'VW_11', 'WW_11', 'PW_11', 'TW_11', 'UP_11', 'VP_11', 'WP_11', 'PP_11', 'TP_11', 
             'UT_11', 'VT_11', 'WT_11', 'PT_11', 'TT_11'
]
lib_terms = np.array(lib_terms)

def constrain_matrix(constraint_tuples):
    n_features = len(lib_terms)
    n_eqns = 5
    n_constraints = len(constraint_tuples)
    constraintlhs = np.zeros((n_constraints, int(n_features*n_eqns)))
    constraintrhs = np.zeros(n_constraints)
    
    #for each constraint
    for ii, tup in enumerate(constraint_tuples):
        term_index = np.where(lib_terms == tup[1])[0]
        constraintlhs[ii, term_index + n_features*tup[0]] = 1 #set the term we want to constrain, in the equation

        if tup[-1] is None:
            pass
        else:
            term_index = np.where(lib_terms == tup[-1])[0]
            constraintlhs[ii, term_index+ n_features*tup[2]] = -1 #constrain coefficients to be equal

    return constraintlhs, constraintrhs

"""
specify a list of tuples [(eqn1, term1, eqn2, term2)]
where eqn1 is the equation you want to constrain, and term1 is the term you want to constraint. 

if you specify (eqn1, term1, None, None), then term1 in equation 1 is automatically set to zero. 
If you specifiy (eqn1, term1, eqn2, term2), then term1 is constrained to equal term2 in eqn2 
"""
# Remove all diffusive terms except U_11, U_22, U_33, V_11, V_22, V_33, W_11, W_22, W_33, T_11, T_22, T_33
nonlinear_diffusive = ['U_23', 'V_23', 'W_23', 'P_23', 'T_23', 'U_13', 'V_13', 'W_13', 'P_13', 'T_13', 
                       'U_12', 'V_12', 'W_12', 'P_12', 'T_12',
                       'UU_33', 'VU_33', 'WU_33', 'PU_33', 'TU_33', 'UV_33', 'VV_33', 'WV_33', 'PV_33', 'TV_33', 
                       'UW_33', 'VW_33', 'WW_33', 'PW_33', 'TW_33', 'UP_33', 'VP_33', 'WP_33', 'PP_33', 'TP_33', 
                       'UT_33', 'VT_33', 'WT_33', 'PT_33', 'TT_33',
                       'UU_23', 'VU_23', 'WU_23', 'PU_23', 'TU_23', 'UV_23', 'VV_23', 'WV_23', 'PV_23', 'TV_23', 
                       'UW_23', 'VW_23', 'WW_23', 'PW_23', 'TW_23', 'UP_23', 'VP_23', 'WP_23', 'PP_23', 'TP_23', 
                       'UT_23', 'VT_23', 'WT_23', 'PT_23', 'TT_23', 
                       'UU_22', 'VU_22', 'WU_22', 'PU_22', 'TU_22', 'UV_22', 'VV_22', 'WV_22', 'PV_22', 'TV_22', 
                       'UW_22', 'VW_22', 'WW_22', 'PW_22', 'TW_22', 'UP_22', 'VP_22', 'WP_22', 'PP_22', 'TP_22', 
                       'UT_22', 'VT_22', 'WT_22', 'PT_22', 'TT_22',
                       'UU_13', 'VU_13', 'WU_13', 'PU_13', 'TU_13', 'UV_13', 'VV_13', 'WV_13', 'PV_13', 'TV_13', 
                       'UW_13', 'VW_13', 'WW_13', 'PW_13', 'TW_13', 'UP_13', 'VP_13', 'WP_13', 'PP_13', 'TP_13', 
                       'UT_13', 'VT_13', 'WT_13', 'PT_13', 'TT_13', 
                       'UU_12', 'VU_12', 'WU_12', 'PU_12', 'TU_12', 'UV_12', 'VV_12', 'WV_12', 'PV_12', 'TV_12', 
                       'UW_12', 'VW_12', 'WW_12', 'PW_12', 'TW_12', 'UP_12', 'VP_12', 'WP_12', 'PP_12', 'TP_12', 
                       'UT_12', 'VT_12', 'WT_12', 'PT_12', 'TT_12', 
                       'UU_11', 'VU_11', 'WU_11', 'PU_11', 'TU_11', 'UV_11', 'VV_11', 'WV_11', 'PV_11', 'TV_11', 
                       'UW_11', 'VW_11', 'WW_11', 'PW_11', 'TW_11', 'UP_11', 'VP_11', 'WP_11', 'PP_11', 'TP_11', 
                       'UT_11', 'VT_11', 'WT_11', 'PT_11', 'TT_11'
]
constraint_tuples = [(j, elem, None, None) for j in [0,1,2,4] for elem in nonlinear_diffusive]

unnecessary_terms = ["U", "V", "W", "P", 
                     "U_3", "V_3", "W_3", "T_3", "P_33", 
                     "U_2", "V_2", "W_2", "T_2", "P_22", 
                     "U_1", "V_1", "W_1", "T_1", "P_11",
                     "VU_1", "WU_1", "PU_1", "TU_1", "UU_2", "WU_2", "PU_2", "TU_2", "UU_3", "VU_3", "PU_3", "TU_3",
                     "VV_1", "WV_1", "PV_1", "TV_1", "UV_2", "WV_2", "PV_2", "TV_2", "UV_3", "VV_3", "PV_3", "TV_3",
                     "VW_1", "WW_1", "PW_1", "TW_1", "UW_2", "WW_2", "PW_2", "TW_2", "UW_3", "VW_3", "PW_3", "TW_3",
                     "VT_1", "WT_1", "PT_1", "TT_1", "UT_2", "WT_2", "PT_2", "TT_2", "UT_3", "VT_3", "PT_3", "TT_3",
                     "UP_1", "VP_1", "WP_1", "PP_1", "TP_1",
                     "UP_2", "VP_2", "WP_2", "PP_2", "TP_2",
                     "UP_3", "VP_3", "WP_3", "PP_3", "TP_3"
]    
constraint_tuples += [(j, elem, None, None) for j in [0,1,2,4] for elem in unnecessary_terms]

# Constraint from incompressibility: Grad U = 0 (U_1=-(V_2 + W_3)
IncomConstr_Ueqn =[(0, 'V_2', None, None), (0, 'UV_2', None, None), (0, 'VV_2', None, None), 
                   (0,'WV_2', None, None), (0, 'PV_2', None, None), (0, 'TV_2', None, None),
                   (0, 'W_3', None, None), (0, 'UW_3', None, None), (0, 'VW_3', None, None), 
                   (0,'WW_3', None, None), (0, 'PW_3', None, None), (0, 'TW_3', None, None)]
constraint_tuples += IncomConstr_Ueqn

# Constraint from incompressibility: Grad U = 0 (V_2=-(U_1 + W_3)
IncomConstr_Veqn =[(1, 'U_1', None, None), (1, 'UU_1', None, None), (1, 'VU_1', None, None), 
                   (1,'WU_1', None, None), (1, 'PU_1', None, None), (1, 'TU_1', None, None),
                   (1, 'W_3', None, None), (1, 'UW_3', None, None), (1, 'VW_3', None, None), 
                   (1,'WW_3', None, None), (1, 'PW_3', None, None), (1, 'TW_3', None, None)]
constraint_tuples += IncomConstr_Veqn

# Constraint from incompressibility: Grad U = 0 (W_3=-(U_1 + V_2)
IncomConstr_Weqn =[(2, 'U_1', None, None), (2, 'UU_1', None, None), (2, 'VU_1', None, None), 
                   (2,'WU_1', None, None), (2, 'PU_1', None, None), (2, 'TU_1', None, None),
                   (2, 'V_2', None, None), (2, 'UV_2', None, None), (2, 'VV_2', None, None), 
                   (2,'WV_2', None, None), (2, 'PV_2', None, None), (2, 'TV_2', None, None)]
constraint_tuples += IncomConstr_Weqn

##### Remove P' equation completely #####
get_rid_of_p = [(3, elem, None, None) for elem in lib_terms]
constraint_tuples += get_rid_of_p

##### Apply vector constraints within equations separately ###### 
constraint_tuples += [(0, "UU_1", 0, "VU_2")] # vector advection constraint
constraint_tuples += [(0, "UU_1", 0, "WU_3")] # vector advection constraint
constraint_tuples += [(0, "U_11", 0, "U_22")] # vector diffusion constraint
constraint_tuples += [(0, "U_11", 0, "U_33")] # vector diffusion constraint
constraint_tuples += [(1, "UV_1", 1, "VV_2")] # vector advection constraint
constraint_tuples += [(1, "UV_1", 1, "WV_3")] # vector advection constraint
constraint_tuples += [(1, "V_11", 1, "V_22")] # vector diffusion constraint
constraint_tuples += [(1, "V_11", 1, "V_33")] # vector diffusion constraint
constraint_tuples += [(2, "UW_1", 2, "VW_2")] # vector advection constraint
constraint_tuples += [(2, "UW_1", 2, "WW_3")] # vector advection constraint
constraint_tuples += [(2, "W_11", 2, "W_22")] # vector diffusion constraint
constraint_tuples += [(2, "W_11", 2, "W_33")] # vector diffusion constraint
constraint_tuples += [(4, "UT_1", 4, "VT_2")] # vector advection constraint
constraint_tuples += [(4, "UT_1", 4, "WT_3")] # vector advection constraint
constraint_tuples += [(4, "T_11", 4, "T_22")] # vector diffusion constraint
constraint_tuples += [(4, "T_11", 4, "T_33")] # vector diffusion constraint

##### Apply vector constraints between equations ###### 
constraint_tuples += [(0, "UU_1", 1, "UV_1")] # vector advection constraint
constraint_tuples += [(0, "UU_1", 2, "UW_1")] # vector advection constraint
constraint_tuples += [(0, "UU_1", 4, "UT_1")] # vector advection constraint
constraint_tuples += [(0, "U_11", 1, "V_11")] # vector diffusion constraint
constraint_tuples += [(0, "U_11", 2, "W_11")] # vector diffusion constraint
constraint_tuples += [(0, "U_11", 4, "T_11")] # vector diffusion constraint
constraint_tuples += [(0,  "P_1", 1,  "P_2")] # vector pressure constraint 
constraint_tuples += [(0,  "P_1", 2,  "P_3")] # vector pressure constraint 

#### Switch off non-U diffusive terms in U' equation ###
constraint_tuples += [(0, "V_11", None, None)]
constraint_tuples += [(0, "V_22", None, None)]
constraint_tuples += [(0, "V_33", None, None)]
constraint_tuples += [(0, "W_11", None, None)]
constraint_tuples += [(0, "W_22", None, None)]
constraint_tuples += [(0, "W_33", None, None)]
constraint_tuples += [(0, "T_11", None, None)]
constraint_tuples += [(0, "T_22", None, None)]
constraint_tuples += [(0, "T_33", None, None)]
#### Switch off non-V diffusive terms in V' equation ###
constraint_tuples += [(1, "U_11", None, None)]
constraint_tuples += [(1, "U_22", None, None)]
constraint_tuples += [(1, "U_33", None, None)]
constraint_tuples += [(1, "W_11", None, None)]
constraint_tuples += [(1, "W_22", None, None)]
constraint_tuples += [(1, "W_33", None, None)]
constraint_tuples += [(1, "T_11", None, None)]
constraint_tuples += [(1, "T_22", None, None)]
constraint_tuples += [(1, "T_33", None, None)]
#### Switch off non-W diffusive terms in W' equation ###
constraint_tuples += [(2, "U_11", None, None)]
constraint_tuples += [(2, "U_22", None, None)]
constraint_tuples += [(2, "U_33", None, None)]
constraint_tuples += [(2, "V_11", None, None)]
constraint_tuples += [(2, "V_22", None, None)]
constraint_tuples += [(2, "V_33", None, None)]
constraint_tuples += [(2, "T_11", None, None)]
constraint_tuples += [(2, "T_22", None, None)]
constraint_tuples += [(2, "T_33", None, None)]
#### Switch off non-T diffusive terms in T' equation ###
constraint_tuples += [(4, "U_11", None, None)]
constraint_tuples += [(4, "U_22", None, None)]
constraint_tuples += [(4, "U_33", None, None)]
constraint_tuples += [(4, "V_11", None, None)]
constraint_tuples += [(4, "V_22", None, None)]
constraint_tuples += [(4, "V_33", None, None)]
constraint_tuples += [(4, "W_11", None, None)]
constraint_tuples += [(4, "W_22", None, None)]
constraint_tuples += [(4, "W_33", None, None)]

#print(constraint_tuples)
constraintlhs, constraintrhs = constrain_matrix(constraint_tuples)

#d=4
#n_features =  int(1 + (1+2*d + d*(1-d)/2)**2) #this is the wrong d 
#print("n_features", n_features) 
#n_targets = 5 #number of equations
#Rayleigh = 1e6
#Prandtl = 1
## Free-fall velocity scaling V0 - (alpha*DeltatT*g*DeltaH)^(1/2)
## Free-fall time-scale T0 (L0/V0) - DeltaH / ((alpha*DeltatT*g*DeltaH)^(1/2))
##
##     alpha and gravitational acceleration
##     - alpha * Theta * g_i  =>> + T * z_unit_vector
##     i.e. alpha=1.0, g_i = -g_z, g=1.0
##
##     Kinematic viscosity
#nu = (Rayleigh / Prandtl) ** (-1/2)
##     Thermal diffusivity
#kappa = (Rayleigh * Prandtl) ** (-1/2)
#
#feature_names = ["R", "R11", "R12", "R22", "Fx", "Fz", "Q", "Ubar", "Wbar", "Temp"] #the state vector, LHS of our problem, the final three are control variables
## For convenience to find the number of libraries I'm using


## Fit the clean data
##optimizer = ps.STLSQ(threshold=5e-6, alpha=1e-12, normalize_columns=False)
#"""
#- I set the threshold small as this controls the regularisation of the coefficients. Since we know what terms we want, thi#s isn't important
#- I decreased the tolerance
#- You might need to increase the number of iterations
#- Nu needs to picked for an appropriate balance between fitting the data and fitting the constraints. 
#"""
u_dot_train_integral = weak_lin_lib.convert_u_dot_integral(u_shaped_noiseless)
#
#
##optimizer = ps.ConstrainedSR3(threshold =0.1, nu = 1.0, tol =1e-12, max_iter = 5000, 
##                              thresholder="l0", verbose=True) #using a very small amount of regularisation as our problem is fully defined
##opt = ps.STLSQ(threshold = 0.000001, alpha=0.000000000001) 
##optimizer = ps.SSR(alpha=0.0001, max_iter = 7, kappa = 0.000001) #could try a greedy method to get the viscous terms? 
##optimizer = ps.FROLS(max_iter=7, alpha=0.1) # i think greedy methods are quite memory intensive depending on how they work
## Frols forms the correlation library and calculating the correlation is probably intense
##MIOSR is a good optimizer typically, but you need a free Gurobi academic license to work it

coeffs = [28]
#coeffs = [5,10,15,20,25]
#coeffs = [10]
msqe = []
msqe2 = []
msqe3 = []

for n_coeff in coeffs:
    print("Model with",n_coeff, "coefficients")
    opt = ps.MIOSR(target_sparsity = n_coeff, alpha=0.00000000000000001,
    constraint_lhs = constraintlhs, constraint_rhs = constraintrhs) #works with small alpha


    model = ps.SINDy(
        feature_library=weak_lin_lib, optimizer=opt, feature_names = ["U", "V", "W", "P", "T"]
    )

    ## get the fitted udot integrals 
    model.fit(u_shaped_noiseless, t=t, unbias=False) #set unbias = False with MIOSR
    model.print(precision=8)
    msqe.append(np.sqrt(
        (
            np.sum((u_dot_train_integral - opt.Theta_ @ opt.coef_.T) ** 2, axis=0)
            / np.sum(u_dot_train_integral**2, axis=0)
        )
        / u_dot_train_integral.shape[0])) #calculate the msqe for each equation
    print("msqe")
    print(msqe[-1])


opt = ps.MIOSR(target_sparsity = None, group_sparsity=(4,4,5,0,3),
               alpha=0.00000000000000001) 

model = ps.SINDy(
    feature_library=weak_lin_lib, optimizer=opt, feature_names = ["U", "V", "W", "P", "T"]
)

print("Model with with group sparsity")
## get the fitted udot integrals 
model.fit(u_shaped_noiseless, t=t, unbias=False) #set unbias = False with MIOSR
model.print(precision=8)

msqe2.append(np.sqrt(
    (
        np.sum((u_dot_train_integral - opt.Theta_ @ opt.coef_.T) ** 2, axis=0)
        / np.sum(u_dot_train_integral**2, axis=0)
    )
    / u_dot_train_integral.shape[0])) #calculate the msqe for each equation
print("msqe2")
print(msqe2[-1])

print("Model with viscous terms")

opt = ps.MIOSR(target_sparsity = None, group_sparsity=(7,7,8,0,6),
               alpha=0.00000000000000001) 

model = ps.SINDy(
    feature_library=weak_lin_lib, optimizer=opt, feature_names = ["U", "V", "W", "P", "T"]
)

print("Model with with viscous terms")
## get the fitted udot integrals 
model.fit(u_shaped_noiseless, t=t, unbias=False) #set unbias = False with MIOSR
model.print(precision=8)

msqe3.append(np.sqrt(
    (
        np.sum((u_dot_train_integral - opt.Theta_ @ opt.coef_.T) ** 2, axis=0)
        / np.sum(u_dot_train_integral**2, axis=0)
    )
    / u_dot_train_integral.shape[0])) #calculate the msqe for each equation
print("msqe3")
print(msqe3[-1])


