"""
Build python script to reproduce Garaud et al. 2010

Usage:
      python garaud_sindy_1.py

Options:

"""
import sys
import h5py
import numpy as np
import matplotlib
#matplotlib.use('Agg')
import matplotlib.pyplot as plt
#from dedalus.extras import plot_tools
from mpl_toolkits.mplot3d import Axes3D
from sklearn.linear_model import Lasso
from scipy.io import loadmat
from sklearn.metrics import mean_squared_error
from scipy.integrate import solve_ivp
from array import array
from STLSQ_functions import *
from scipy.interpolate import interp2d 

import pysindy as ps

# Ignore matplotlib deprecation warnings
import warnings
warnings.filterwarnings("ignore", category=UserWarning)

# Check if the script is called with the expected number of arguments
if len(sys.argv) != 4:
    print("Usage: python script.py integerlist: 1.xdivisions 2.zdivisions 3.tdivisions")
    sys.exit(1)

"""
Points with the code:
    1. We should try to understand how the size of the integration domain impacts the results. 
What is important in smaller domains vs bigger domains? Is a larger domain integral always better? Is it only better for large scale terms?
    2. How do we choose the time averaging, what if we do almost no time averaging? 
    3. Why is the equation in W not working as well? 
    4. How will we identify the viscous terms? I think either normalisations or a different optimiser
    5. If we get this to work, can we find the mean terms and remove them in the identification process? is this possible?
    6. Why are we not seeing viscous dissipation terms in the temperature equation even though they should be order 1?
    7. This problem seems to perform poorly with high L2 regularisation for some reason 

    8. It would be great if we could justify the flow being incompressible and thus not needing a pressure evolution equation. 
"""

# Seed the random number generators for reproducibility
np.random.seed(100)

integrator_keywords = {}
integrator_keywords['rtol'] = 1e-12
integrator_keywords['method'] = 'LSODA'
integrator_keywords['atol'] = 1e-12

####
#file_range = [1,2,3] # TOO FEW!
#file_range = [1,2,3,4,5]
file_range = [1,2,3,4,5,6,7,8]
#file_range = [1,2,3,4,5,6,7,8,9,10] #np.arange(1,2) # set file range for HDF5 files
#file_range = [1,2,3,4,5,6,7,8,9,10,11,12,13,14,15]
#file_range = [1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20] 
#file_range = [1,2,3,4,5,10,11,12,13,14,15,21,22,23,24,25,31,32,33,34,35,41,42,43,44,45] 
#file_range = [1,2,3,4,5,10,11,12,13,14,15,21,22,23,24,25,31,32,33,34,35,41,42,43,44,45] 
####
time = np.array([]) # read in the simulation time
nx, nz = 1024, 384
nt = int(50*file_range[-1])

Pres_store = np.zeros((nt, nx, nz), dtype=np.float64)
Temp_store = np.zeros((nt, nx, nz), dtype=np.float64)
U_store = np.zeros((nt, nx, nz), dtype=np.float64)
W_store = np.zeros((nt, nx, nz), dtype=np.float64)
#time_store = np.zeros((nt, nx, nz), dtype=np.float64)

for qq, ii in enumerate(file_range):
#    with h5py.File('snapshots/snapshots_s{}.h5'.format(ii)) as file:
#    with h5py.File('/nobackup/matcjw/dedalus/2dconvection/F-C-Ra-10^10/2023-11-17-restart/snapshots/snapshots_s{}.h5'.format(ii)) as file:
#    with h5py.File('/nobackup/matcjw/dedalus/2dconvection/new-F-C-Ra-10^6/restart/snapshots/snapshots_s{}.h5'.format(ii)) as file:
    with h5py.File('/nobackup/matcjw/dedalus/2dconvection/new-F-C-Ra-10^8/restart/snapshots/snapshots_s{}.h5'.format(ii)) as file:
        print("Doing file", ii)
        if ii==1:
            print(file["scales"].keys())
            #xdomain = file["scales/x_hash_89db91f00173784cfec0de1995a005933e22284d"][:]
            #zdomain = file["scales/z_hash_3c859c36a9c2f295aaab1dec0ec24d51221a3714"][:]            
            xdomain = np.array(file["tasks/pressure"].dims[1][0])
            zdomain = np.array(file["tasks/pressure"].dims[2][0])
        # Time stamps
        time = np.append(time, file["scales/sim_time"][:])
        
        print(len(time))
        # Full data
        if qq+50<nt:
            Pres_store[qq*50:(qq+1)*50,:,:] = file['tasks/pressure'][:,:,:]  # Pressure
            Temp_store[qq*50:(qq+1)*50,:,:] = file['tasks/buoyancy'][:,:,:]  # Temperature / buoyancy
            U_store[qq*50:(qq+1)*50,:,:] = file['tasks/u_x'][:,:,:]          # u_x; x-velocity
            W_store[qq*50:(qq+1)*50,:,:] = file['tasks/u_z'][:,:,:]         # u_z; z-velocity
        else:
            Pres_store[qq*50:,:,:] = file['tasks/pressure'][:,:,:]  # Pressure
            Temp_store[qq*50:,:,:] = file['tasks/buoyancy'][:,:,:]  # Temperature / buoyancy
            U_store[qq*50:,:,:] = file['tasks/u_x'][:,:,:]          # u_x; x-velocity
            W_store[qq*50:,:,:] = file['tasks/u_z'][:,:,:]         # u_z; z-velocity

#plt.figure(figsize=(15,10))
#plt.plot(zdomain, np.ones(zdomain.shape), "b.")
#plt.xlim((0,0.5))
#plt.savefig("chebyshev_domain.pdf")

# Transpose axes into what SINDy expects (time,x,z) -> (x,z,time)
Pres = np.transpose(Pres_store, (1,2,0))
Temp = np.transpose(Temp_store, (1,2,0))
U = np.transpose(U_store, (1,2,0))
W = np.transpose(W_store, (1,2,0))

del Pres_store, Temp_store, U_store, W_store

# Note: we now have the necessary data in the right form for PDEFIND vis SINDy

#print(xdomain[500],xdomain1[500])
#print(zdomain[190],zdomain1[190])
#print(zdomain)

# Print confirmation data
print('Temp has shape ',Temp.shape,' and size ',Temp.size)

#print(xdomain.shape)
#print(zdomain.shape)
#print(Pres[:,:,0].shape)

# interpolation from Chebyshev to Fourier
print(len(time))
new_z = np.linspace(0,1,385)[:-1] # Uniform z space
new_Pres = np.zeros((nx, nz, nt), dtype=np.float64)
new_Temp = np.zeros((nx, nz, nt), dtype=np.float64)
new_U = np.zeros((nx, nz, nt), dtype=np.float64)
new_W = np.zeros((nx, nz, nt), dtype=np.float64)
for ii in range(len(time)):
    # interpolate
    interp_p = interp2d(np.squeeze(xdomain),np.squeeze(zdomain),Pres[:,:,ii].T)
    new_Pres[:,:,ii] = interp_p(np.squeeze(xdomain),new_z).T
    interp_t = interp2d(np.squeeze(xdomain),np.squeeze(zdomain),Temp[:,:,ii].T)
    new_Temp[:,:,ii] = interp_t(np.squeeze(xdomain),new_z).T
    interp_U = interp2d(np.squeeze(xdomain),np.squeeze(zdomain),U[:,:,ii].T)
    new_U[:,:,ii] = interp_U(np.squeeze(xdomain),new_z).T
    interp_W = interp2d(np.squeeze(xdomain),np.squeeze(zdomain),W[:,:,ii].T)
    new_W[:,:,ii] = interp_W(np.squeeze(xdomain),new_z).T

#print(new_Pres.shape)
#print(new_Temp.shape)
#print(new_U.shape)
#print(new_W.shape)

# Test - interpolation results seems OK!
#print('Old: ',xdomain[500],zdomain[190],Pres[500,190,5])
#print('Old: ',xdomain[500],zdomain[191],Pres[500,191,5])
#print('New: ',xdomain[500],new_z[190],new_Pres[500,190,5])
#exit()

# Use interpolated data
Pres = new_Pres
Temp = new_Temp
U = new_U
W = new_W

# Use new uniform z domain
zdomain = new_z

#%% I'll try to keep the convention similar to the compressible example (this isn't that much data)
#u_shaped_noiseless = np.zeros((Ubar.shape[0], Ubar.shape[1], 7)) # 1D z,t

u_shaped_noiseless = np.zeros((U.shape[0], U.shape[1], U.shape[2], 4))
u_shaped_noiseless[:,:,:,0] = U
u_shaped_noiseless[:,:,:,1] = W
u_shaped_noiseless[:,:,:,2] = Pres
u_shaped_noiseless[:,:,:,3] = Temp

# the control variables are essentially terms we want included in the library but do not have independent equations

print("u data shape", u_shaped_noiseless.shape)
truncate_time = None #index to truncate at

#u_shaped_noiseless = u_shaped_noiseless[:,:truncate_time, :] #truncate the data to make it slightly easier to work with 
u_shaped_noiseless = u_shaped_noiseless[:,:,:truncate_time, :] #truncate the data to make it slightly easier to work with 

time = time[:truncate_time]
print(u_shaped_noiseless.shape)

######## truncate spatially ##########
#x_1, x_2 = 256, 512 #1
#x_1, x_2 = 0, 1024 #2 Full Grid
x_1, x_2 = 256, 768 #3
#x_1, x_2 = 0, None
#x_1, x_2 = 665, 921 # captures considerable dynamics

xdomain = xdomain[x_1:x_2] #take half of the domain
u_shaped_noiseless = u_shaped_noiseless[x_1:x_2, :,:] 

#z_1, z_2 = 100, 280 #A 100 away from boundaries - bulk flow.
#z_1, z_2 = 50, 334 #B
#z_1, z_2 = 0, 15 #C - picks up diffusive terms
#z_1, z_2 = 333, 383 #D
#z_1, z_2 = 10, 100 # tends not to pick up diffusive terms at higher Ra
z_1, z_2 = 0, None # Full grid

zdomain = zdomain[z_1:z_2]
u_shaped_noiseless = u_shaped_noiseless[:,z_1:z_2,:]
print("u shaped noiseless", u_shaped_noiseless.shape)

# Opt 1+A Files 1-15 works pretty well for Navier-Stokes
# Opt 3+C Files 1-15 works well for some diffusive terms
#####################################

#%% 
# Generate time array - should this run from 50 or zero? 
# A: it doesn't matter really, the time stepping is the relevant part
#but I will start from t=0
nt = u_shaped_noiseless.shape[2] # (nx,nz,time,variable)
#t = np.squeeze(np.linspace(start_time, end_time, nt, endpoint=True)) #this assumes uniform stepping, which isn't normally the case with Dedalus code
#dt = t[1] - t[0]
#print(t)
t = time #change to the output of Dedalus
t -= t[0] #change to be from zero
start_time, end_time = t[0], t[-1]
print('Start Time:',time[0])
print('End time:  ',time[-1])

# X grid. Fourier grid in realspace linearly spaced from zero.
nx = u_shaped_noiseless.shape[0] # (nx,nz,time,variable)
Lx0 = xdomain[0]
Lx = xdomain[-1] 
x = xdomain #np.linspace(0, Lx, nx, endpoint=False)
#dx = x[1] - x[0]
print('X-grid:',x[0],x[-1])

# Z grid. Fourier grid in realspace linearly spaced from zero.
nz = u_shaped_noiseless.shape[1] # (nx,nz,time,variable)
Lz0 = zdomain[0]
Lz = zdomain[-1] # Full domain was 2, but we removed haslf the data at read in
z = zdomain[:u_shaped_noiseless.shape[1]] #index the original z domain to the truncation
#dz = z[1] - z[0]
print('Z-grid:',z[0],z[-1])

# Question: are these grid points the left edge of cells? DO we want centre of cells (add dx/z or dz/2)?

#
spatiotemporal_grid = np.zeros((nx, nz,  nt, 3))

spatiotemporal_grid[:,:,:,0] = x[:,np.newaxis, np.newaxis] 
spatiotemporal_grid[:, :,:,1] = z[np.newaxis, :, np.newaxis]
spatiotemporal_grid[:, :,:, 2] = t[np.newaxis, np.newaxis, :]

X, T = np.meshgrid(z, t)
XT = spatiotemporal_grid

######################### GENERAL PDE LIBRARY TERMS ########################################
## body NS terms
#Model = int(sys.argv[1])
divisions_x = int(sys.argv[1])
divisions_z = int(sys.argv[2])
divisions_t = int(sys.argv[3])
#divisions_x = 10#10 # i think something in x is causing a problem, variables and derivatives in z on their own
# seem to be ok 
#divisions_z = 3
#divisions_t = 10 # 10

K = 300

print(' ')
print(' ')
print('Divisions_x: ',divisions_x)
print('Divisions_z: ',divisions_z)
print('Divisions_t: ',divisions_t)

weak_lin_lib = ps.WeakPDELibrary(
    library_functions=[lambda x: x],
    function_names = [lambda x: x],
    derivative_order=2, 
    spatiotemporal_grid=XT,
    include_interaction = True,
    periodic = False,   
    p=6, # Our highest order term is second derivatives so we need to make sure this is set right
    #uniform=False, there's some argument we should check is false
    K=K, #number of sub-domains to evaluate the integral on, essentially an ensembling process
    H_xt=[(Lx-Lx0) / divisions_x, (Lz-Lz0) / divisions_z, (end_time-start_time) / divisions_t] #size of the sub domains, not sure how to choose this
)

#print("Created the weak pde library") 

#%%

################ Constrain each equation so only the correct terms appear ############
"""
The constraints have to be of the size (n_constraints, n_features*n_targets)

this means we have to have a row for every term for every each in the system

In our case n_constraints will be the sum of the number of terms in the feature library - the number of terms with unknown coefficients for each equation

"""
lib_terms = ['U', 'W', 'P', 'T', 'U_2', 'W_2', 'P_2', 'T_2', 'U_22', 'W_22', 'P_22', 'T_22', 
             'U_1', 'W_1', 'P_1', 'T_1', 'U_12', 'W_12', 'P_12', 'T_12', 'U_11', 'W_11', 'P_11', 'T_11', 
             'UU_2', 'WU_2', 'PU_2', 
             'TU_2', 'UW_2', 'WW_2', 'PW_2', 'TW_2', 'UP_2', 'WP_2', 'PP_2', 'TP_2', 'UT_2', 'WT_2', 
             'PT_2', 'TT_2', 'UU_22', 'WU_22', 
             'PU_22', 'TU_22', 'UW_22', 'WW_22', 'PW_22', 'TW_22', 'UP_22', 'WP_22', 'PP_22', 'TP_22', 
             'UT_22', 'WT_22', 'PT_22', 'TT_22', 
             'UU_1', 'WU_1', 'PU_1', 'TU_1', 'UW_1', 'WW_1', 'PW_1', 'TW_1', 'UP_1', 'WP_1', 'PP_1', 
             'TP_1', 'UT_1', 'WT_1', 'PT_1', 'TT_1', 
             'UU_12', 'WU_12', 'PU_12', 'TU_12', 'UW_12', 'WW_12', 'PW_12', 'TW_12', 'UP_12', 'WP_12', 
             'PP_12', 'TP_12', 'UT_12', 'WT_12', 
             'PT_12', 'TT_12', 'UU_11', 'WU_11', 'PU_11', 'TU_11', 'UW_11', 'WW_11', 'PW_11', 'TW_11', 
             'UP_11', 'WP_11', 'PP_11', 'TP_11',
             'UT_11', 'WT_11', 'PT_11', 'TT_11']
lib_terms = np.array(lib_terms)


def constrain_matrix(constraint_tuples):
    n_features = len(lib_terms)
    n_eqns = 4
    n_constraints = len(constraint_tuples)
    constraintlhs =np.zeros((n_constraints, int(n_features*n_eqns)))
    constraintrhs = np.zeros(n_constraints)
    
    #for each constraint
    for ii, tup in enumerate(constraint_tuples):
        term_index = np.where(lib_terms == tup[1])[0]
        constraintlhs[ii, term_index + n_features*tup[0]] = 1 #set the term we want to constrain, in the equation
        
        if tup[-1] is None:
            pass
        else:
            term_index = np.where(lib_terms == tup[-1])[0]
            constraintlhs[ii, term_index+ n_features*tup[2]] = -1 #constrain coefficients to be equal
            
    return constraintlhs, constraintrhs

"""
specify a list of tuples [(eqn1, term1, eqn2, term2)]
where eqn1 is the equation you want to constrain, and term1 is the term you want to constraint. 

if you specify (eqn1, term1, None, None), then term1 in equation 1 is automatically set to zero. 
If you specifiy (eqn1, term1, eqn2, term2), then term1 is constrained to equal term2 in eqn2 
"""
# Remove all diffusive terms except U_11, U_22, W_11, W_22, T_11, T_22 i.e. remove 54 of 60.
nonlinear_diffusive = ['P_22', 'P_12', 'T_12', 'P_11', 'UU_22','WU_22','PU_22','TU_22',
                       'UW_22','WW_22','PW_22','TW_22','UP_22','WP_22','PP_22','TP_22','UT_22','WT_22',
                       'PT_22','TT_22', 
                       'UP_12','WP_12','PP_12','TP_12','UT_12','WT_12','PT_12','TT_12','UU_11','WU_11', 
                       'PU_11','TU_11','UW_11','WW_11','PW_11','TW_11','UP_11','WP_11','PP_11','TP_11',
                       'UT_11','WT_11','PT_11','TT_11']
constraint_tuples = [(j, elem, None, None) for j in [0,1,3] for elem in nonlinear_diffusive]
#unnecessary_terms = ["U", "W", "P", "U_2", "W_2", "T_2", "P_22", "U_1", "W_1", "T_1", 
#                     "U_12", "W_12", "P_12", "T_12", "P_11", "UU_2", "PU_2", "TU_2",
#                     "UW_2", "PW_2", "TW_2", "UP_2", "WP_2", "PP_2", "TP_2",
#                     "UT_2", "PT_2", "TT_2", "WT_1", "PT_1", "TT_1",
#                     "WU_1", "PU_1", "TU_1", "UP_1", "WP_1", "PP_1", "TP_1",
#                     "WW_1", "PW_1", "TW_1"]    
#constraint_tuples += [(j, elem, None, None) for j in [0,1,3] for elem in unnecessary_terms]
#constraint_tuples +=[(3, 'TT_11', None, None), (3, 'TT_22', None, None), #set all these terms to zero
#                     (3, 'PT_11', None, None), (3, 'PT_22', None, None),
#                     (3, 'WT_11', None, None), (3, 'WT_22', None, None),
#                     (3, 'UT_11', None, None), (3, 'UT_22', None, None),
#                     (3, 'TP_11', None, None), (3, 'TP_22', None, None),
#                     (3, 'UP_11', None, None), (3, 'WP_11', None, None),
#                     (3, 'UP_22', None, None), (3, 'WP_22', None, None),
#                     ]
# Constraint from incompressibility: Grad U = 0 (U_1=-W_2)
IncomConstr_Ueqn =[(0, 'W_2', None, None), (0, 'UW_2', None, None), (0, 'WW_2', None, None), 
                   (0, 'PW_2', None, None), (0, 'TW_2', None, None),
                   (0,'W_12', None, None), (0,'UW_12', None, None), (0,'WW_12', None, None),
                   (0,'PW_12', None, None), (0,'TW_12', None, None),
                   (0,'U_12', None, None), (0,'UU_12', None, None), (0,'WU_12', None, None),
                   (0,'PU_12', None, None), (0,'TU_12', None, None)]
constraint_tuples += IncomConstr_Ueqn
IncomConstr_Weqn =[(1, 'U_1', None, None), (1, 'UU_1', None, None), (1, 'WU_1', None, None), 
                   (1, 'PU_1', None, None), (1, 'TU_1', None, None),
                   (1,'U_12', None, None), (1,'UU_12', None, None), (1,'WU_12', None, None),
                   (1,'PU_12', None, None), (1,'TU_12', None, None),
                   (1,'W_12', None, None), (1,'UW_12', None, None), (1,'WW_12', None, None),
                   (1,'PW_12', None, None), (1,'TW_12', None, None)]
constraint_tuples += IncomConstr_Weqn
get_rid_of_p = [(2, elem, None, None) for elem in lib_terms]
constraint_tuples += get_rid_of_p 
constraint_tuples += [(0, "UU_1", 0, "WU_2")] # symmetry constraint
constraint_tuples += [(0, "UU_1", 1, "UW_1")] # vector equation constraint
constraint_tuples += [(0, "UU_1", 1, "WW_2")] # vector equation constraint
constraint_tuples += [(0, "U_11", 0, "U_22")] # symmetry constraint
constraint_tuples += [(0, "U_11", 1, "W_11")] # vector equation constraint
constraint_tuples += [(0, "U_22", 1, "W_22")] # vector equation constraint
constraint_tuples += [(0,  "P_1", 1,  "P_2")] # vector equation constraint 
constraint_tuples += [(1, "UW_1", 1, "WW_2")] # symmetry constraint
constraint_tuples += [(1, "W_11", 1, "W_22")] # symmetry constraint
constraint_tuples += [(3, "UT_1", 3, "WT_2")] # symmetry constraint
constraint_tuples += [(3, "T_11", 3, "T_22")] # symmetry constraint
constraint_tuples += [(0, "W_11", None, None)]
constraint_tuples += [(0, "W_22", None, None)]
constraint_tuples += [(0, "T_11", None, None)]
constraint_tuples += [(0, "T_22", None, None)]
constraint_tuples += [(1, "U_11", None, None)]
constraint_tuples += [(1, "U_22", None, None)]
constraint_tuples += [(1, "T_11", None, None)]
constraint_tuples += [(1, "T_22", None, None)]
constraint_tuples += [(3, "U_11", None, None)]
constraint_tuples += [(3, "U_22", None, None)]
constraint_tuples += [(3, "W_11", None, None)]
constraint_tuples += [(3, "W_22", None, None)]
#print(constraint_tuples)
constraintlhs, constraintrhs = constrain_matrix(constraint_tuples)

d=4
n_features =  int(1 + (1+2*d + d*(1-d)/2)**2) #this is the wrong d 
#print("n_features", n_features) 
n_targets = 4 #number of equations
Rayleigh = 1e10
Prandtl = 1
# Free-fall velocity scaling V0 - (alpha*DeltatT*g*DeltaH)^(1/2)
# Free-fall time-scale T0 (L0/V0) - DeltaH / ((alpha*DeltatT*g*DeltaH)^(1/2))
#
#     alpha and gravitational acceleration
#     - alpha * Theta * g_i  =>> + T * z_unit_vector
#     i.e. alpha=1.0, g_i = -g_z, g=1.0
#
#     Kinematic viscosity
nu = (Rayleigh / Prandtl) ** (-1/2)
#     Thermal diffusivity
kappa = (Rayleigh * Prandtl) ** (-1/2)
            


feature_names = ["R", "R11", "R12", "R22", "Fx", "Fz", "Q", "Ubar", "Wbar", "Temp"] #the state vector, LHS of our problem, the final three are control variables
# For convenience to find the number of libraries I'm using
            

# Fit the clean data
#optimizer = ps.STLSQ(threshold=5e-6, alpha=1e-12, normalize_columns=False)
"""
- I set the threshold small as this controls the regularisation of the coefficients. Since we know what terms we want, this isn't important
- I decreased the tolerance
- You might need to increase the number of iterations
- Nu needs to picked for an appropriate balance between fitting the data and fitting the constraints. 
"""
u_dot_train_integral = weak_lin_lib.convert_u_dot_integral(u_shaped_noiseless)


optimizer = ps.ConstrainedSR3(threshold =0.1, nu = 1.0, tol =1e-12, max_iter = 5000, 
                              thresholder="l0", verbose=True) #using a very small amount of regularisation as our problem is fully defined
opt = ps.STLSQ(threshold = 0.000001, alpha=0.000000000001) 
#optimizer = ps.SSR(alpha=0.0001, max_iter = 7, kappa = 0.000001) #could try a greedy method to get the viscous terms? 
#optimizer = ps.FROLS(max_iter=7, alpha=0.1) # i think greedy methods are quite memory intensive depending on how they work
# Frols forms the correlation library and calculating the correlation is probably intense
#MIOSR is a good optimizer typically, but you need a free Gurobi academic license to work it

coeffs = [15]
#coeffs = [5,10,15,20,25]
#coeffs = [10]
msqe = []
msqe2 = []
msqe3 = []

for n_coeff in coeffs:
    #print("Model with",n_coeff, "coefficients")
    opt = ps.MIOSR(target_sparsity = n_coeff, alpha=0.00000000000000001,
                   constraint_lhs = constraintlhs, constraint_rhs = constraintrhs) #works with small alpha
    
                
    model = ps.SINDy(
        feature_library=weak_lin_lib, optimizer=opt, feature_names = ["U", "W", "P", "T"]
    )
                
    ## get the fitted udot integrals 
    model.fit(u_shaped_noiseless, t=t, unbias=False) #set unbias = False with MIOSR
    model.print(precision=8)
    msqe.append(np.sqrt(
        (
            np.sum((u_dot_train_integral - opt.Theta_ @ opt.coef_.T) ** 2, axis=0)
            / np.sum(u_dot_train_integral**2, axis=0)
        )
        / u_dot_train_integral.shape[0])) #calculate the msqe for each equation
    print("msqe")
    print(msqe[-1])
                
print("total msqe")
print(msqe)

#######################################
print("Model with with group sparsity")

opt = ps.MIOSR(target_sparsity = None, group_sparsity=(3,4,0,2),
               alpha=0.00000000000000001) 
    
model = ps.SINDy(
    feature_library=weak_lin_lib, optimizer=opt, feature_names = ["U", "W", "P", "T"]
)
            
## get the fitted udot integrals 
model.fit(u_shaped_noiseless, t=t, unbias=False) #set unbias = False with MIOSR
model.print(precision=8)
msqe2.append(np.sqrt(
    (
        np.sum((u_dot_train_integral - opt.Theta_ @ opt.coef_.T) ** 2, axis=0)
        / np.sum(u_dot_train_integral**2, axis=0)
    )
    / u_dot_train_integral.shape[0])) #calculate the msqe for each equation
print("msqe2")
print(msqe2[-1])
    
######################################
print("Model with with viscous terms")

opt = ps.MIOSR(target_sparsity = None, group_sparsity=(5,6,0,4),
               alpha=0.00000000000000001) 
    
model = ps.SINDy(
    feature_library=weak_lin_lib, optimizer=opt, feature_names = ["U", "W", "P", "T"]
)

## get the fitted udot integrals 
model.fit(u_shaped_noiseless, t=t, unbias=False) #set unbias = False with MIOSR
model.print(precision=8)
msqe3.append(np.sqrt(
    (
        np.sum((u_dot_train_integral - opt.Theta_ @ opt.coef_.T) ** 2, axis=0)
        / np.sum(u_dot_train_integral**2, axis=0)
    )
    / u_dot_train_integral.shape[0])) #calculate the msqe for each equation
print("msqe3")
print(msqe3[-1])

with open('fulltest-reduced2.txt', 'a+') as file:
    # Write variables to the file on the same line
    file.write(f'{divisions_x} {divisions_z} {divisions_t} {msqe[-1]} {msqe3[-1]}\n')

#print("Im tryig to get the viscous terms at the end but it's not working right now")
#udot_integral_prediction = opt.Theta_ @ opt.coef_.T
#print("THeta and opt shapes", opt.Theta_.shape, opt.coef_.T.shape)
#print(udot_integral_prediction.shape, u_dot_train_integral.shape) 

#u_dot_train_integral -= udot_integral_prediction 
            
#Xi, coefhist, coeffs = get_coef(u_dot_train_integral, opt.Theta_, model.get_feature_names(), initial_guess=None, unbias=True, maxiter=10,
#                               thresholds = 0.00000001*np.ones(4), alpha =0.000000000000001)



#model.coefficients()[:] = coeffs
#model.print(precision =8)
# we have small terms appearing so make sure to print a high enough precision


#%% 
#print("R2 score: ", model.score(u_shaped_noiseless, t=t))

